@@ -1,7 +1,7 @@
 /*
 	mmv
 
-	Copyright (c) 2021-2023 Reuben Thomas.
+	Copyright (c) 2021-2024 Reuben Thomas.
 	Copyright (c) 1990 Vladimir Lanin.
 
 	This program is distributed under the GNU GPL version 3, or, at your
@@ -50,16 +50,11 @@
 #include <dirent.h>
 #include <limits.h>
 
-#include <gc/gc.h>
-
 #include "progname.h"
 #include "binary-io.h"
 #include "dirname.h"
 #include "pathmax.h"
 #include "xalloc.h"
-#ifndef _WIN32
-#include "ignore-value.h"
-#endif
 
 #include "cmdline.h"
 
@@ -78,13 +73,12 @@
 #define OVERWRITE 0x004
 #define NORMMOVE 0x008
 #define XMOVE 0x010
-#define DIRMOVE 0x020
 #define APPEND 0x040
 #define HARDLINK 0x100
 #define SYMLINK 0x200
 
 #define COPY (NORMCOPY | OVERWRITE)
-#define MOVE (NORMMOVE | XMOVE | DIRMOVE)
+#define MOVE (NORMMOVE | XMOVE)
 #define LINK (HARDLINK | SYMLINK)
 
 static char COPYNAME[] = "mcp";
@@ -145,7 +139,6 @@ typedef struct {
 	char h_err;
 } HANDLE;
 
-#define R_ISX 0x01
 #define R_SKIP 0x02
 #define R_DELOK 0x04
 #define R_ISALIASED 0x08
@@ -173,7 +166,6 @@ typedef struct {
 
 
 static int op, badstyle, delstyle, verbose, noex, matchall, mkdirs;
-static int patflags;
 
 static size_t ndirs = 0, dirroom;
 static size_t ndirs_nonexistent = 0, dirroom_nonexistent;
@@ -208,7 +200,7 @@ static REP mistake;
 
 static const char *home;
 static size_t homelen;
-static uid_t uid, euid;
+static uid_t uid;
 static mode_t oldumask;
 static ino_t cwdd = (ino_t)-1L;
 static dev_t cwdv = (dev_t)-1L;
@@ -481,7 +473,6 @@ static FILEINFO *fsearch(char *s, DIRINFO *d)
 	FILEINFO *f = (FILEINFO *)xmalloc(sizeof(FILEINFO));
 	f->fi_name = s;
 	FILEINFO **res = bsearch(&f, d->di_fils, d->di_nfils, sizeof(FILEINFO *), fcmp);
-	free(f);
 	return res != NULL ? *res : NULL;
 }
 
@@ -568,7 +559,6 @@ static int make_path(char *path)
 				}
 				if (fail) {
 					umask(oldumask);
-					free(npath);
 					return(1);
 				}
 			}
@@ -576,7 +566,6 @@ static int make_path(char *path)
 				if (verbose)
 					printf("cannot create directory `%s': %s", npath, strerror(errno));
 				umask(oldumask);
-				free(npath);
 				return(1);
 			}
 		}
@@ -587,10 +576,26 @@ static int make_path(char *path)
 	}
 
 	umask(oldumask);
-	free(npath);
 	return(0);
 }
 
+/* Create a non-existent directory and update its DIRINFO. */
+static int make_directory(HANDLE *h) {
+	int res = make_path(h->h_name);
+	if (res != 0)
+		fprintf(stderr, "Strange, couldn't create directory %s.\n",  h->h_name);
+	else {
+		struct stat dstat;
+		if (stat(h->h_name, &dstat) || (dstat.st_mode & S_IFMT) != S_IFDIR) {
+			fprintf(stderr, "Strange, couldn't stat new directory %s.\n", h->h_name);
+			res = -1;
+		}
+		h->h_di->di_vid = dstat.st_dev;
+		h->h_di->di_did = dstat.st_ino;
+	}
+	return res;
+}
+
 static HANDLE *hadd(char *n)
 {
 	if (nhandles == handleroom)
@@ -747,64 +752,47 @@ static HANDLE *checkdir(char *p, char *pathend, int makedirs)
 	return(h);
 }
 
-static int checkto(HANDLE *hfrom, char *f, HANDLE **phto, char **pnto, FILEINFO **pfdel)
+static int checkto(char *f, HANDLE **phto, char **pnto, FILEINFO **pfdel)
 {
 	char tpath[PATH_MAX + 1];
-	char *pathend;
 	FILEINFO *fdel = NULL;
-	size_t hlen, tlen;
 
-	if (op & DIRMOVE) {
-		*phto = hfrom;
-		hlen = strlen(hfrom->h_name);
-		pathend = fullrep + hlen;
-		memmove(pathend, fullrep, strlen(fullrep) + 1);
-		memmove(fullrep, hfrom->h_name, hlen);
-		if ((fdel = *pfdel = fsearch(pathend, hfrom->h_di)) != NULL) {
-			*pnto = xstrdup(fdel->fi_name);
-			getstat(fullrep, fdel);
-		}
-		else
-			*pnto = xstrdup(pathend);
-	}
-	else {
-		pathend = getpath(tpath);
-		hlen = (size_t)(pathend - fullrep);
+	char *pathend = getpath(tpath);
+	size_t hlen = (size_t)(pathend - fullrep);
+	*phto = checkdir(tpath, tpath + hlen, mkdirs);
+	if (
+	    *phto != NULL &&
+	    *pathend != '\0' &&
+	    (fdel = *pfdel = fsearch(pathend, (*phto)->h_di)) != NULL &&
+	    (getstat(fullrep, fdel), fdel->fi_stflags & FI_ISDIR) &&
+	    (strcmp(pathend, fullrep) != 0)
+	    ) {
+		size_t tlen = strlen(pathend);
+		strcpy(pathend + tlen, SLASHSTR);
+		tlen++;
+		strcpy(tpath + hlen, pathend);
+		pathend += tlen;
+		hlen += tlen;
 		*phto = checkdir(tpath, tpath + hlen, mkdirs);
-		if (
-			*phto != NULL &&
-			*pathend != '\0' &&
-			(fdel = *pfdel = fsearch(pathend, (*phto)->h_di)) != NULL &&
-			(getstat(fullrep, fdel), fdel->fi_stflags & FI_ISDIR) &&
-			(strcmp(pathend, fullrep) != 0)
-		) {
-			tlen = strlen(pathend);
-			strcpy(pathend + tlen, SLASHSTR);
-			tlen++;
-			strcpy(tpath + hlen, pathend);
-			pathend += tlen;
-			hlen += tlen;
-			*phto = checkdir(tpath, tpath + hlen, mkdirs);
-		}
-
-		if (*pathend == '\0') {
-			*pnto = xstrdup(f);
-			if ((size_t)(pathend - fullrep) + strlen(f) >= PATH_MAX) {
-				strcpy(fullrep, TOOLONG);
-				return(-1);
-			}
-			strcat(pathend, f);
-			if (*phto != NULL) {
-				fdel = *pfdel = fsearch(f, (*phto)->h_di);
-				if (fdel != NULL)
-					getstat(fullrep, fdel);
-			}
-		}
-		else if (fdel != NULL)
-			*pnto = fdel->fi_name;
-		else
-			*pnto = xstrdup(pathend);
 	}
+
+	if (*pathend == '\0') {
+		*pnto = xstrdup(f);
+		if ((size_t)(pathend - fullrep) + strlen(f) >= PATH_MAX) {
+			strcpy(fullrep, TOOLONG);
+			return(-1);
+		}
+		strcat(pathend, f);
+		if (*phto != NULL) {
+			fdel = *pfdel = fsearch(f, (*phto)->h_di);
+			if (fdel != NULL)
+				getstat(fullrep, fdel);
+		}
+	}
+	else if (fdel != NULL)
+		*pnto = fdel->fi_name;
+	else
+		*pnto = xstrdup(pathend);
 	return(0);
 }
 
@@ -877,19 +865,19 @@ static int badrep(HANDLE *hfrom, FILEINFO *ffrom, HANDLE **phto, char **pnto, FI
 		!(op & SYMLINK)
 	)
 		printf("%s -> %s : . and .. can't be renamed.\n", pathbuf, fullrep);
-	else if (repbad || checkto(hfrom, f, phto, pnto, pfdel) || badname(*pnto))
+	else if (repbad || checkto(f, phto, pnto, pfdel) || badname(*pnto))
 		printf("%s -> %s : bad new name.\n", pathbuf, fullrep);
 	else if (*phto == NULL)
 		printf("%s -> %s : %s.\n", pathbuf, fullrep,
 			direrr == H_NOREADDIR ?
 			"no read or search permission for target directory" :
-			"target directory does not exist");
+			"target directory does not exist (you could use --makedirs)");
 	else if (!dwritable(*phto))
 		printf("%s -> %s : no write permission for target directory.\n",
 			pathbuf, fullrep);
 	else if (
 		(*phto)->h_di->di_vid != hfrom->h_di->di_vid &&
-		(*pflags = R_ISX, (op & (NORMMOVE | HARDLINK)))
+		(op & (NORMMOVE | HARDLINK))
 	)
 		printf("%s -> %s : cross-device move.\n",
 			pathbuf, fullrep);
@@ -1078,7 +1066,7 @@ static int dostage(char *lastend, char *pathend, char **start1, size_t *len1, in
 					(*pf)->fi_rep = MISTAKE;
 				} else {
 					(*pf)->fi_rep = p = (REP *)xmalloc(sizeof(REP));
-					p->r_flags = flags | patflags;
+					p->r_flags = flags;
 					p->r_hfrom = h;
 					p->r_ffrom = *pf;
 					p->r_hto = hto;
@@ -1214,11 +1202,6 @@ static int parsepat(void)
 	for (p = lastname; (c = *p) != '\0'; p++)
 		switch (c) {
 		case SLASH:
-			if (op & DIRMOVE) {
-				printf("%s -> %s : no path allowed in target under -r.\n",
-					from, to);
-				return(-1);
-			}
 			lastname = p + 1;
 			break;
 		case '#':
@@ -1628,9 +1611,8 @@ static void doreps(void)
 			if (mkdirs && p->r_hto->h_di->di_flags & DI_NONEXISTENT) {
 				if (verbose)
 					printf("creating directory %s\n", p->r_hto->h_name);
-				int res = make_path(p->r_hto->h_name);
-				if (res != 0)
-					fprintf(stderr, "Strange, couldn't create directory %s.\n",  p->r_hto->h_name);
+				// FIXME: check the return value.
+				make_directory(p->r_hto);
 				p->r_hto->h_di->di_flags &= ~DI_NONEXISTENT;
 			}
 			strcat(fullrep, p->r_nto);
@@ -1658,7 +1640,7 @@ static void doreps(void)
 					(op & SYMLINK) ?
 						symlink((p->r_flags & R_ONEDIRLINK) ? fstart : pathbuf,
 							fullrep) :
-					p->r_flags & R_ISX ?
+					p->r_hto->h_di->di_vid != p->r_hfrom->h_di->di_vid ?
 						copymove(p) :
 					/* move */
 						rename(pathbuf, fullrep)
@@ -1705,7 +1687,6 @@ int main(int argc, char *argv[])
 	}
 	oldumask = umask(0);
 #ifndef _WIN32
-	euid = geteuid();
 	uid = getuid();
 #endif
 	signal(SIGINT, breakout);
@@ -1740,12 +1721,11 @@ int main(int argc, char *argv[])
 	else if (args_info.terminate_given != 0)
 		badstyle = ABORTBAD;
 
-	if (args_info.move_given != 0)
+	// The hidden --rename/-r option is recognised for backwards compatibility.
+	if (args_info.move_given != 0 || args_info.rename_given != 0)
 		op = NORMMOVE;
 	else if (args_info.copydel_given != 0)
 		op = XMOVE;
-	else if (args_info.rename_given != 0)
-		op = DIRMOVE;
 	else if (args_info.copy_given != 0)
 		op = NORMCOPY;
 	else if (args_info.overwrite_given != 0)
@@ -1757,7 +1737,7 @@ int main(int argc, char *argv[])
 	else if (args_info.symlink_given != 0)
 		op = SYMLINK;
 	else {
-		const char *name = basename(program_name);
+		const char *name = base_name(program_name);
 
 		if (strcmp(name, COPYNAME) == 0)
 			op = NORMCOPY;
@@ -1769,14 +1749,6 @@ int main(int argc, char *argv[])
 			op = XMOVE;
 	}
 
-#ifndef _WIN32
-	if (euid != uid && !(op & DIRMOVE)) {
-		/* Best effort. */
-		ignore_value(setuid(uid));
-		ignore_value(setgid(getgid()));
-	}
-#endif
-
 	if (badstyle != ASKBAD && delstyle == ASKDEL)
 		delstyle = NODEL;
 
@@ -1803,7 +1775,5 @@ int main(int argc, char *argv[])
 		scandeletes(skipdel);
 	doreps();
 
-	cmdline_parser_free(&args_info);
-
 	return(failed ? 2 : nreps == 0 && (paterr || badreps));
 }
